import acm.program.*;
import acm.util.*;
import java.util.*;
import java.io.*;

public class MachineLearning extends ConsoleProgram {
	/* The learning rate to use in perceptron learning. */
	private static final double LEARNING_RATE = 1.0e-5;
	
	/* The fraction of the data we can misclassify and still call the result
	 * "good enough."
	 */
	private static final double DESIRED_ACCURACY = 5.0e-3;
	
	public void run() {
		setFont("DejaVuSerif-BOLD-24");
		
		/* Load a training set of images and train the perceptrons on those images. */
		ArrayList<LabeledImage> trainingData = loadImages("training-data");
		ArrayList<double[]> perceptrons = trainPerceptrons(trainingData);
		
		/* Load a testing set of images and test the perceptrons on those images. */
		ArrayList<LabeledImage> testingData = loadImages("testing-data");
		testPerceptrons(perceptrons, testingData);
		visualizePerceptrons(perceptrons);
	}
	
	/**
	 * Given a list of perceptrons and a test data set, tests the perceptrons on
	 * those images and reports the accuracy.
	 * 
	 * @param perceptrons The perceptrons to test.
	 * @param testingData The data set on which they should be tested.
	 */
	private void testPerceptrons(ArrayList<double[]> perceptrons,
			                     ArrayList<LabeledImage> testingData) {
		/* Track how many times the perceptrons were correct. */
		int numRight = 0;
		
		for (LabeledImage image: testingData) {
			double[] pixels = image.getPixels();
			
			/* Determine what we think the image is by storing our best guess
			 * and how confident we are.
			 */
			int bestGuess = 0;
			double bestConfidence = Double.NEGATIVE_INFINITY;
			
			/* Try out each perceptron and see how it does. */
			for (int digit = 0; digit < perceptrons.size(); digit++) {
				double confidence = dotProduct(perceptrons.get(digit), pixels);
				
				if (confidence > bestConfidence) {
					bestGuess = digit;
					bestConfidence = confidence;
				}
			}
			
			/* Mark if we were correct. */
			if (bestGuess == image.getLabel()) {
				numRight++;
			}
		}
		
		println("Accuracy: " + numRight * 100.0 / testingData.size() + "%.");
	}
	
	/**
	 * Trains a group of perceptrons to recognize the digits specified by the
	 * trainingData set.
	 * 
	 * @param trainingData The training data to use.
	 * @return A list of perceptrons trained to recognize them.
	 */
	private ArrayList<double[]> trainPerceptrons(ArrayList<LabeledImage> trainingData) {
		ArrayList<double[]> perceptrons = new ArrayList<double[]>();
		
		/* Train a perceptron for each digit. */
		for (int digit = 0; digit < 10; digit++) {
			print("Training a perceptron to recognize " + digit + "...");
			perceptrons.add(trainPerceptron(trainingData, digit));
			println("   done!");
		}
		return perceptrons;
	}
	
	/**
	 * Given two arrays of doubles of the same length, computes their dot product
	 * (the sum of their pairwise products).
	 * 
	 * @param first The first array.
	 * @param second The second array.
	 * @return Their dot product.
	 */
	private double dotProduct(double[] first, double[] second) {
		double result = 0.0;
		for (int i = 0; i < first.length; i++) {
			result += first[i] * second[i];
		}
		return result;
	}
	
	/**
	 * Returns whether a collection of errors is "good enough" for the perceptron
	 * to stop learning.
	 * 
	 * @param errors The list of errors made.
	 * @return Whether there are sufficiently few of them.
	 */
	private boolean goodEnough(int[] errors) {
		/* Count how many errors we have.  An error is when the difference
		 * between the real and actual values is nonzero.
		 */
		int numWrong = 0;
		for (int error: errors) {
			if (error != 0) {
				numWrong++;
			}
		}
		return (double)numWrong / errors.length <= DESIRED_ACCURACY;
	}
	
	/**
	 * Trains a perceptron to recognize the given digit.
	 * 
	 * @param trainingData The training set to use.
	 * @param digit The digit that should be learned.
	 * @return A perceptron to recognize that digit.
	 */
	private double[] trainPerceptron(ArrayList<LabeledImage> trainingData,
			                         int digit) {
		/* Create a weights vector with space for all pixels, plus one for the +1
		 * constant term.
		 */
		double[] weights = new double[LabeledImage.IMAGE_SIZE + 1];
		
		while (true) {
			/* Check how our perceptron is doing right now. */
			int[] errors = new int[trainingData.size()];
			for (int i = 0; i < trainingData.size(); i++) {
				LabeledImage image = trainingData.get(i);
				double[] pixels = image.getPixels();
				
				/* Determine our guess and the actual answer, then store the difference. */
				int ourGuess = (dotProduct(weights, pixels) > 0)? 1 : 0;
				int actual   = (image.getLabel() == digit)? 1 : 0;
				
				errors[i] = actual - ourGuess;
			}
			
			/* If the perceptron has converged, return the current weights. */
			if (goodEnough(errors))
				return weights;
			
			/* Adjust the weights based on the formula from the slides. */
			for (int i = 0; i < trainingData.size(); i++) {
				LabeledImage image = trainingData.get(i);
				double[] pixels = image.getPixels();
				
				for (int p = 0; p < pixels.length; p++) {
					weights[p] += LEARNING_RATE * errors[i] * pixels[p];
				}
			}
		}
	}
	
	/************* Starter Code Below This Point *************/
	
	/**
	 * Loads all of the images from the specified data set,
	 * returning them in an ArrayList.
	 * 
	 * @param filename The name of the file from which data should be read.
	 * @return An ArrayList of all the labeled images.
	 */
	private ArrayList<LabeledImage> loadImages(String filename) {
		ArrayList<LabeledImage> result = new ArrayList<LabeledImage>();
		
		try {
			/* Open the file for reading. */
			BufferedReader br = new BufferedReader(new FileReader(filename));
			
			while (true) {
				/* Pull all of the image data, then the label. */
				String data = br.readLine();
				String label = br.readLine();
				
				/* Skip a blank line. */
				br.readLine();
				
				/* Stop if no data remains. */
				if (label == null) break;
				
				/* Construct a new LabeledImage and add it to the
				 * result list.
				 */
				result.add(new LabeledImage(data, Integer.parseInt(label)));
			}
			
			br.close();
			return result;
			
		} catch (IOException e) {
			throw new ErrorException("Error reading file.");
		}
	}
	
	/**
	 * Creates windows that visualize each of the perceptrons.
	 * 
	 * @param perceptrons The perceptrons to visualize.
	 */
	private void visualizePerceptrons(ArrayList<double[]> perceptrons) {
		/* Iterate in reverse order so the windows show up in the correct order. */
		for (int digit = perceptrons.size() - 1; digit >= 0; digit--) {
			new PerceptronWindow(perceptrons.get(digit), digit).setVisible(true);
		}
	}
}
